/*
 * Interface to an Ericsson 388 at the frame protocol level
 *
 * $Id: frame.c 1.1 1997/12/22 19:58:20 dds Exp $

 *
 * (C) 1997 Diomidis D. Spinellis.  All rights reserved.
 *
 * Permission to use, copy, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 * 
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include <stdio.h>
#include <bios.h>
#include <time.h>
#include "sio.h"

static int portno = 2;
static int verbose = 1;

typedef unsigned char uchar;

enum e_state {
	IDLE_600,	/* Idle at 600 baud */
	IDLE_9600,	/* Idle at 9600 baud */
	WAIT_ACK,	/* Command sent, waiting for ACK */
	WAIT_LEN,	/* Start of frame seen, waiting for frame len */
	FRAME_DATA,	/* Frame len received, processing frame data */
} state;

static uchar framedata[1024];
static int framelen, frameindex;
static int have_frame;

/*
 * Process an incoming character
 */
static void
process(uchar c)
{
	if (c == 0xff)
		return;
	if (verbose > 1)
		printf("process: char = 0x%02x\n", c);
	switch (state) {
	case IDLE_9600:
		if (c == 0x2) {
			state = WAIT_LEN;
			if (verbose > 1)
				printf("process: frame start\n", c);
		} else {
			if (verbose > 0)
				printf("process: bad char (not SOF) 0x%02x\n", c);
		}
		break;
	case WAIT_ACK:
		if (c == 0x6) {
			state = IDLE_9600;
			if (verbose > 1)
				printf("process: got ACK\n", c);
		} else {
			if (verbose > 0)
				printf("process: bad char (not ACK) 0x%02x\n", c);
		}
		break;
	case WAIT_LEN:
		if (verbose > 1)
			printf("process: got len (%d)\n", c);
		state = FRAME_DATA;
		framelen = c;
		frameindex = 0;
		break;
	case FRAME_DATA:
		if (verbose > 1)
			printf("process: got frame byte (0x%02x)\n", c);
		framedata[frameindex++] = c;
		if (frameindex == framelen) {
			sio_put(portno, 0x6);		/* ACK */
			state = IDLE_9600;
			if (verbose > 1)
				printf("process: finished frame\n", c);
			have_frame = 1;
		}
		break;
	default:
		printf("process: unknown state (%d)\n", state);
		sio_close();
		exit(1);
	}
}

/*
 * Expect a specific character.
 * Return 0 if ok, -1 if timeout, -2 if a different character is received.
 */
static int
expect(int val)
{
	time_t start, now;
	int c;

	if (verbose > 1)
		printf("expect(0x%02x)\n", val);
	time(&start);
	for (;;) {
		c = sio_get(portno);
		if (c == -1) {
			time(&now);
			if (now - start > 3) {
				if (verbose > 0)
					printf("expect(0x%02x): timeout\n", val);
				return (-1);
			}
		} else if (c == val) {
			if (verbose > 1)
				printf("expect(0x%02x): ok\n", val);
			return (0);
		} else {
			if (verbose > 0)
				printf("expect(0x%02x): got 0x%02x\n", c);
			return (-1);
		}
	}
}

/*
 * Sleep >= n seconds
 */
static int
sleep(int val)
{
	time_t start, now;

	time(&start);
	do {
		time(&now);
	} while (now - start < val);
}

/*
 * Resynchronise to 9600 baud
 */
static int
resync(void)
{
	int ret;
	int i;

	if (verbose > 0)
		printf("resync: synchronizing communication...\n");
	/* Send a 9600 command that we will not ACK */
	sio_put(portno, 0x02);
	sio_put(portno, 0x02);
	sio_put(portno, 0x05);
	sio_put(portno, 0x60);
	sleep(3);
	if (verbose > 1)
		printf("resync: flushing input buffer\n");
	for (i = 0; i < 100; i++)
		sio_get(portno);
	sleep(2);
	if (verbose > 1)
		printf("resync: switching to 600 baud\n");
	sio_init(portno, 600, 8, 1, SIO_NO_PARITY);
	state = IDLE_600;
	if (verbose > 1)
		printf("resync: switched to 600 baud\n");
	sleep(2);
	sio_put(portno, 0xa);
	sio_put(portno, 0xa);
	sleep(2);
	for (i = 0; i < 100; i++)
		sio_get(portno);
	sio_put(portno, 0xa);
	if ((ret = expect(0xa)) != 0) {
		if (verbose > 1)
			printf("resync: first sync failed.  Retrying.\n");
		for (i = 0; i < 100; i++)
			sio_get(portno);
		sio_put(portno, 0xa);
		if ((ret = expect(0xa)) != 0)
			return (-1);
	}
	if ((ret = expect(0xae)) != 0)
		return (-1);
	if ((ret = expect(0xa3)) != 0)
		return (-1);
	sio_put(portno, 0x30);
	sio_put(portno, 0x08);
	sleep(2);
	sio_init(portno, 9600, 8, 1, SIO_NO_PARITY);
	state = IDLE_9600;
	sleep(2);
	if (verbose > 0)
		printf("resync: ok\n");
	return (0);
}

/*
 * Send a specific command
 */
static void
sendcmd(uchar *cmd, int len)
{
	int i;

	if (state != IDLE_9600) {
		printf("sendcmd: bad state (%d)\n", state);
		return;
	}
	sio_put(portno, 0x2);
	sio_put(portno, (uchar)len);
	for (i = 0; i < len; i++)
		sio_put(portno, cmd[i]);
	state = WAIT_ACK;
	have_frame = 0;
	if (verbose > 1)
		printf("Command sent\n");
}

static int phonedump = 0;

/*
 * Dump frame characters
 */
static void
dump_frame(void)
{
	int i;

	if (framedata[0] == 0x08 && framedata[1] == 0x00) {
		printf("Phone ID: ");
		for (i = 2; i < framelen; i++)
			putchar(framedata[i]);
		putchar('\n');
	} else if (framedata[0] == 0x06 && framedata[1] == 0x60) {
		printf("Battery level (0-64): %d\n", framedata[2]);
	} else if (framedata[0] == 0x06 && framedata[1] == 0x20) {
		printf("Reception level (0-48): %d\n", framedata[2]);
	} else if (framedata[0] == 0x46 && framedata[1] == 0x01) {
		char name[40], phone[20], number[5];

		sscanf(framedata + 2, "%[^;];%[^;];%[^;];", number, phone, name);
		if (phonedump) {
			fprintf(stderr, "%s\t%s\t%s\n", number, name, phone);
			printf("%s\t%s\t%s\n", number, name, phone);
		} else
			printf("Memory:\t%s\nName:\t%s\nPhone:\t%s\n", number, name, phone);
	} else {
		printf("Unknown frame (0x%02x, 0x%02x):", framedata[0], framedata[1]);
		for (i = 2; i < framelen; i++)
			printf(" %02x", framedata[i]);
		printf(" - [");
		for (i = 2; i < framelen; i++)
			if (framedata[i] >= ' ' && framedata[i] <= '~')
				putchar(framedata[i]);
			else
				putchar('.');
		printf("]\n");
	}
}

static void
usage(void)
{
	printf("usage: frame -v n -p n\n"
		"\t-v n\tSet verbosity level to n (0-2; default 1)\n"
		"\t-p n\tSet serial port to n (default 2)\n"
		"\t-d\tDump phonebook contents\n"
	);
	exit(1);
}


main(int argc, char *argv[])
{
	int ret, n;
	uchar c;
	uchar cmd[64];

	char buff[100];

	extern char *optarg;

	while ((ret=getopt(argc,argv,"dp:v:?")) != EOF)
		switch ((char)ret) {
		case 'p':
			portno = atoi(optarg);
			break;
		case 'v':
			verbose = atoi(optarg);
			break;
		case 'd':
			phonedump = 1;
			break;
		case '?':
			usage();
			break;
		}
	sio_init(portno, 9600, 8, 1, SIO_NO_PARITY);
	ret = resync();
	if (ret != 0) {
		printf("resync failed\n");
		sio_close();
		exit(1);
	}

	if (phonedump) {
		int waitframes;

		memcpy(cmd, "\x45\x03\000", 3);
		sendcmd(cmd, 3);
		waitframes = 3;
		for (;;) {
			c = sio_get(portno);
			process(c);
			if (have_frame) {
				dump_frame();
				waitframes--;
				have_frame = 0;
			}
			if (state == IDLE_9600 && waitframes == 0) {
				if (phonedump == 100) {
					sio_close();
					return (0);
				}
				cmd[2] = phonedump++;
				sendcmd(cmd, 3);
				waitframes = 1;
			}
		}
	}

	printf("Ready for command input\n");
	for (;;) {
		c = sio_get(portno);
		process(c);
		if (have_frame) {
			dump_frame();
			have_frame = 0;
		}
		if (_bios_keybrd(_KEYBRD_READY)) {
			c = getch();
			switch (c) {
			default:
			case '?':		/* Help */
				printf("Available commands are:\n"
					"i	Phone ID\n"
					"b	Battery level\n"
					"r	Reception level\n"
					"p	Phonebook contents\n"
					"v	Voice call\n"
					"ESC	Quit\n"
				);
				break;
			case 'i':		/* Ask for phone ID */
				sendcmd("\x7\x0", 2);		
				break;
			case 'b':		/* Battery level */
				sendcmd("\x5\x60", 2);		
				break;
			case 'r':		/* Reception level */
				sendcmd("\x5\x20", 2);		
				break;
			case 'p':		/* Phonebook contents */
				printf("Enter memory number:");
				gets(buff);
				sscanf(buff, " %d ", &n);
				memcpy(cmd, "\x45\x03", 2);
				cmd[2] = (uchar)n;
				sendcmd(cmd, 3);
				break;
			case 'v':		/* Voice call */
				printf("Enter phone number:");
				memcpy(cmd, "\x21\x01", 2);
				gets(buff);
				memcpy(cmd + 2, buff, strlen(buff));
				cmd[strlen(buff) + 2] = ';';
				sendcmd(cmd, strlen(buff) + 3);
				break;
			case 'q':
			case '\x1b':
				sio_close();
				return (0);
			}
		}
	}
}
