/*
 * Serial input output library
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>
#include <conio.h>
#include "sio.h"

#define TRUE 1
#define FALSE 0


/*------------------------------------------------------------------*
	  This program bypasses the less-than-adequate PC BIOS
	  communications routines and installs a serial interrupt
	  handler. Direct access to PC hardware allows the program to
	  run at faster baud rates and eliminates the need for
	  the main program to continuously poll the serial port for
	  data; thus implementing background communications. Data that
	  enters the serial port is stored in a circular buffer.
  
		* Compile this program with Test Stack Overflow OFF.
  
	The following are application level routines for:
  
		a. initializing a serial port			sio_init
		b. closing both serial ports			sio_close
		c. reading a character from the receive buffer  sio_get
		d. outputting characters from a serial port	sio_put
*/

#define COM1	1
#define COM2	2
#define COM1BASE        0x3F8	/* Base port address for COM1  */
#define COM2BASE        0x2F8	/* Base port address for COM2  */

/*-
    The 8250 UART has 10 registers accessible through 7 port addresses.
    Here are their addresses relative to COM1BASE and COM2BASE. Note
    that the baud rate registers, (DLL) and (DLH) are active only when
    the Divisor-Latch Access-Bit (DLAB) is on. The (DLAB) is bit 7 of
    the (LCR).
  
	o TXR Output data to the serial port.
	o RXR Input data from the serial port.
	o LCR Initialize the serial port.
	o IER Controls interrupt generation.
	o IIR Identifies interrupts.
	o MCR Send contorl signals to the modem.
	o LSR Monitor the status of the serial port.
	o MSR Receive status of the modem.
	o DLL Low byte of baud rate divisor.
	o DHH High byte of baud rate divisor.
*/
#define TXR             0	/* Transmit register (WRITE) */
#define RXR             0	/* Receive register  (READ)  */
#define IER             1	/* Interrupt Enable          */
#define IIR             2	/* Interrupt ID              */
#define LCR             3	/* Line control              */
#define MCR             4	/* Modem control             */
#define LSR             5	/* Line Status               */
#define MSR             6	/* Modem Status              */
#define DLL             0	/* Divisor Latch Low         */
#define DLH             1	/* Divisor latch High        */


/*-------------------------------------------------------------------*
  Bit values held in the Line Control Register (LCR).
	bit		meaning
	---		-------
	0-1		00=5 bits, 01=6 bits, 10=7 bits, 11=8 bits.
	2		Stop bits.
	3		0=parity off, 1=parity on.
	4		0=parity odd, 1=parity even.
	5		Sticky parity.
	6		Set break.
	7		Toggle port addresses.
 *-------------------------------------------------------------------*/
#define NO_PARITY       0x00
#define EVEN_PARITY     0x18
#define ODD_PARITY      0x08



/*-------------------------------------------------------------------*
  Bit values held in the Line Status Register (LSR).
	bit		meaning
	---		-------
	0		Data ready.
	1		Overrun error - Data register overwritten.
	2		Parity error - bad transmission.
	3		Framing error - No stop bit was found.
	4		Break detect - End to transmission requested.
	5		Transmitter holding register is empty.
	6		Transmitter shift register is empty.
	7               Time out - off line.
 *-------------------------------------------------------------------*/
#define RCVRDY          0x01
#define OVRERR          0x02
#define PRTYERR         0x04
#define FRMERR          0x08
#define BRKERR          0x10
#define XMTRDY          0x20
#define XMTRSR          0x40
#define TIMEOUT		0x80

/*-------------------------------------------------------------------*
  Bit values held in the Modem Output Control Register (MCR).
	bit     	meaning
	---		-------
	0		Data Terminal Ready. Computer ready to go.
	1		Request To Send. Computer wants to send data.
	2		auxillary output #1.
	3		auxillary output #2.(Note: This bit must be
			set to allow the communications card to send
			interrupts to the system)
	4		UART ouput looped back as input.
	5-7		not used.
 *------------------------------------------------------------------*/
#define DTR             0x01
#define RTS             0x02
#define MC_INT		0x08


/*------------------------------------------------------------------*
  Bit values held in the Modem Input Status Register (MSR).
	bit		meaning
	---		-------
	0		delta Clear To Send.
	1		delta Data Set Ready.
	2		delta Ring Indicator.
	3		delta Data Carrier Detect.
	4		Clear To Send.
	5		Data Set Ready.
	6		Ring Indicator.
	7		Data Carrier Detect.
 *------------------------------------------------------------------*/
#define CTS             0x10
#define DSR             0x20


/*------------------------------------------------------------------*
  Bit values held in the Interrupt Enable Register (IER).
	bit		meaning
	---		-------
	0		Interrupt when data received.
	1		Interrupt when transmitter holding reg. empty.
	2		Interrupt when data reception error.
	3		Interrupt when change in modem status register.
	4-7		Not used.
 *------------------------------------------------------------------*/
#define RX_INT          0x01
#define TX_INT		0x02

/*------------------------------------------------------------------*
  Bit values held in the Interrupt Identification Register (IIR).
	bit		meaning
	---		-------
	0		Interrupt pending
	1-2             Interrupt ID code
			00=Change in modem status register,
			01=Transmitter holding register empty,
			10=Data received,
			11=reception error, or break encountered.
	3-7		Not used.
 *------------------------------------------------------------------*/
#define TX_ID		0x02
#define RX_ID           0x04
#define RX_MASK         0x07


/*
 * These are the port addresses of the 8259 Programmable Interrupt Controller
 * (PIC). 
 */
#define IMR             0x21	/* Interrupt Mask Register port */
#define ICR             0x20	/* Interrupt Control Port       */


/*
 * An end of interrupt needs to be sent to the Control Port of the 8259 when
 * a hardware interrupt ends. 
 */
#define EOI             0x20	/* End Of Interrupt */


/*
 * The (IMR) tells the (PIC) to service an interrupt only if it is not masked
 * (FALSE). 
 */
#define IRQ3            0xF7	/* COM2 */
#define IRQ4            0xEF	/* COM1 */


static void interrupt com_int_1(void);
static void interrupt com_int_2(void);
static void setvects(int port);
static void resvects(void);
static void i_enable(int pnum);
static void i_disable(void);
static void comm_on(int pnum);
static void comm_off(void);
static void initserial(int port);
static int SetPort(int Port);
static int SetSpeed(int port, int Speed);
static int SetOthers(int port, int Parity, int Bits, int StopBit);
static int SetSerial(int Port, int Speed, int Parity, int Bits, int StopBit);

#define NOERROR         0	/* No error               */
#define BUFOVFL         1	/* Buffer overflowed      */

#define ESC             0x1B	/* ASCII Escape character */
#define ASCII           0x007F	/* Mask ASCII characters  */
#define SBUFSIZ         0x4000	/* Serial buffer size     */

static int      tsip_port = 1;
static int      port_initialized[2] = {FALSE, FALSE};
static int      io_buf_code[2];
int             portbase[2] = {0, 2};
void            (_interrupt* oldvects[2]) ();

/***************************************************************************/

/* 
 * Input/output queues:
 *
 * The queue checks and operations must be performed with interrupts
 * disabled to guarentee consistency and atomicity.
 */

#define MAXMSG		1024

struct s_siobuf {
	short           last;	/* location of last byte in buffer 		 */
	short           first;	/* location of first byte in buffer 	 */
	byte            msgbuf[MAXMSG + 1];	/* receive buffer							  	 */
} sibuf[2], sobuf[2];

#define q_canwrite(s) ((s->last != (s->first - 1)) && ((s->first) || (s->last != MAXMSG)))
#define q_canread(s) (s->last != s->first)

void 
sio_init(int port, int speed, int bits, int stopbits, int parity)
{
	/* Communications parameters */

	if (SetSerial(port, speed, parity, bits, stopbits) != 0) {
		fprintf(stderr, "Serial Port setup error.\n");
		exit(1);
	}
	initserial(port);

}

void 
sio_close(void)
{
	comm_off();
	resvects();
}


int 
sio_get(int port)
{
	struct s_siobuf   *s;
	int          	   retval = -1;

	s = &sibuf[0];
	if (port == COM2)
		s++;

	_disable();
	if (q_canread(s)) {
		retval = s->msgbuf[s->first++];
		if (s->first > MAXMSG)
			s->first = 0;
	}
	_enable();
	return (retval);

}				/* get_char */

/* Output a character to the serial port */
int 
sio_put(int port, byte x)
{
	int             useport;
	int		retval = 0;

	useport = portbase[port - 1];

	outp(useport + MCR, MC_INT | DTR | RTS);

	/* Wait for Clear To Send from modem */
	/*
	 * while ((inp(useport + MSR) & CTS) == 0) if (!(--timeout)) return
	 * (-1); 
	 */

	/* See if transmitter is clear */
	if (inp(useport + LSR) & XMTRDY) {
		_disable();
		outp(useport + TXR, x);
		_enable();
	} else  {
		struct s_siobuf        *s;

		s = &sobuf[port - 1];
		_disable();
		if (q_canwrite(s)) {
			s->msgbuf[s->last++] = x;
			if (s->last > MAXMSG)
				s->last = 0;
		} else
			retval = -1;
		_enable();
	}
	return (retval);
}

/* Handle communications interrupts */
static void
inthandler(int port)
{
	byte            	ch;
	struct s_siobuf        *s;

	switch (inp(portbase[port] + IIR) & RX_MASK) {
	case RX_ID:
		ch = inp(portbase[port] + RXR);
		s = &sibuf[port];
		if (q_canwrite(s)) {
			s->msgbuf[s->last++] = ch;
			if (s->last > MAXMSG)
				s->last = 0;
		}
		break;
	case TX_ID:
		s = &sobuf[port];
		if (q_canread(s)) {
			ch = s->msgbuf[s->first++];
			if (s->first > MAXMSG)
				s->first = 0;
			outp(portbase[port] + TXR, ch);
		}
		break;
	}
	/* Signal end of hardware interrupt */
	outp(ICR, EOI);
	_enable();
}

static void _interrupt 
com_int_1(void)
{
	inthandler(0);
}

static void _interrupt 
com_int_2(void)
{
	inthandler(1);
}


/* Install our functions to handle communications */
static void 
setvects(int port)
{
	if (port == COM1 && !oldvects[1]) {
		oldvects[1] = _dos_getvect(0x0C);
		_dos_setvect(0x0C, com_int_1);
	} else if (port == COM2 && !oldvects[0]) {
		oldvects[0] = _dos_getvect(0x0B);
		_dos_setvect(0x0B, com_int_2);
	}
}

/* Uninstall our vectors before exiting the program */
static void 
resvects(void)
{
	if (oldvects[0])
		_dos_setvect(0x0B, oldvects[0]);
	if (oldvects[1])
		_dos_setvect(0x0C, oldvects[1]);
}

/* Turn on communications interrupts */
static void 
i_enable(int pnum)
{
	int             c;

	_disable();
	c = inp(portbase[pnum - 1] + MCR) | MC_INT;
	outp(portbase[pnum - 1] + MCR, c);
	outp(portbase[pnum - 1] + IER, RX_INT | TX_INT);
	c = inp(IMR) & (pnum == COM1 ? IRQ4 : IRQ3);
	outp(IMR, c);
	_enable();
}

/* Turn off communications interrupts */
static void 
i_disable(void)
{
	int             c;

	_disable();
	c = inp(IMR) | ~IRQ3 | ~IRQ4;
	outp(IMR, c);

	outp(portbase[0] + IER, 0);
	c = inp(portbase[0] + MCR) & ~MC_INT;
	outp(portbase[0] + MCR, c);

	outp(portbase[1] + IER, 0);
	c = inp(portbase[1] + MCR) & ~MC_INT;
	outp(portbase[1] + MCR, c);

	_enable();
}

/* Tell modem that we're ready to go */
static void 
comm_on(int pnum)
{
	int             c;

	i_enable(pnum);

	c = inp(portbase[pnum - 1] + MCR) | DTR | RTS;
	/* c = inp(portbase[pnum-1] + MCR); */
	outp(portbase[pnum - 1] + MCR, c);
}

/* Go off-line */
static void 
comm_off(void)
{
	i_disable();
	outp(portbase[0] + MCR, 0);
	outp(portbase[1] + MCR, 0);
}

static void 
initserial(int port)
{
	setvects(port);
	comm_on(port);
}


/* Set the port number to use */
static int 
SetPort(int Port)
{
	int             Offset, far * RS232_Addr;

	switch (Port) {		/* Sort out the base address */
	case COM1:
		Offset = 0x0000;
		break;
	case COM2:
		Offset = 0x0002;
		break;
	default:
		return (-1);
	}

	RS232_Addr = (int _far *) (0x00400000l | Offset);	/* Find out where the
								 * port is. */
	if (*RS232_Addr == 0)
		return (-1);	/* If NULL then port not used. */
	portbase[Port - 1] = *RS232_Addr;	/* Otherwise set portbase      */

	return (0);
}

/* This routine sets the speed; will accept funny baud rates. */
/* Setting the speed requires that the DLAB be set on.        */
static int 
SetSpeed(int port, int Speed)
{
	byte            c;
	int             divisor;

	if (Speed == 0)		/* Avoid divide by zero */
		return (-1);
	else
		divisor = (int) (115200L / Speed);

	if (portbase[port - 1] == 0)
		return (-1);

	_disable();
	c = inp(portbase[port - 1] + LCR);
	outp(portbase[port - 1] + LCR, (c | 0x80));	/* Set DLAB */
	outp(portbase[port - 1] + DLL, (divisor & 0x00FF));
	outp(portbase[port - 1] + DLH, ((divisor >> 8) & 0x00FF));
	outp(portbase[port - 1] + LCR, c);	/* Reset DLAB */
	_enable();

	return (0);
}

/* Set other communications parameters */
static int 
SetOthers(int port, int Parity, int Bits, int StopBit)
{
	int             setting;

	if (portbase[port - 1] == 0)
		return (-1);
	if (Bits < 5 || Bits > 8)
		return (-1);
	if (StopBit != 1 && StopBit != 2)
		return (-1);
	if (Parity != NO_PARITY && Parity != ODD_PARITY && Parity != EVEN_PARITY)
		return (-1);

	setting = Bits - 5;
	setting |= ((StopBit == 1) ? 0x00 : 0x04);
	setting |= Parity;

	_disable();
	outp(portbase[port - 1] + LCR, setting);
	_enable();

	return (0);
}

/* Set up the port */
static int 
SetSerial(int Port, int Speed, int Parity, int Bits, int StopBit)
{
	if (SetPort(Port))
		return (-1);
	if (SetSpeed(Port, Speed))
		return (-1);
	if (SetOthers(Port, Parity, Bits, StopBit))
		return (-1);

	return (0);
}
